/**
  *  \file client/chartloc.cc
  *  \brief GChartLocation class
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chartloc.h"
#include "game/univ.h"
#include "game/global.h"
#include "game/ship.h"
#include "game/planet.h"

enum { OBJS_PER_PAGE = 15 };

static GUfo*
findUfoOtherEnd(GPoint pt)
{
    const GUfoType& ufos = getDisplayedTurn().getCurrentUniverse().ty_ufos;
    for (GUfoType::const_iterator it = ufos.begin(), end = ufos.end(); it != end; ++it) {
        if (distanceSquared(it->getPos(), pt) <= it->getRadiusSquared()) {
            if (GUfo* otherEnd = it->getOtherEnd()) {
                if (otherEnd->getPos() != pt) {
                    return otherEnd;
                }
            }
        }
    }
    return 0;
}

/** Constructor. */
GChartLocation::GChartLocation()
    : location(0, 0),
      wrap_adjust(0, 0),
      pid(0),
      view_name(""),
      keymap(0),
      changed(false)
{
    viewport.center.x = viewport.center.y = 0;
}

/** Destructor. */
GChartLocation::~GChartLocation()
{ }

/** Update location. Called when position (location) changed. */
void
GChartLocation::updateLocation()
{
    GObjectReference ref = getCurrentObject();
    GPoint realLocation = location.getCanonicalLocation();

    // Build object lists
    list.clear();
    if (haveDisplayedTurn()) {
        GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
        pid = univ.getPlanetAt(realLocation);
        list.addObjectsAt(univ, realLocation, ls_IncludeForeignShips + ls_IncludePlanet, 0);
        list.postprocessUser(false);

        if (!ref.isValid() && univ.ty_any_ships.isValidIndex(viewport.vec_id)) {
            ref = GObjectReference(univ.ty_any_ships, viewport.vec_id);
        }
    }

    // Reset list index.
    if (list.size() > 0) {
        int n = list.getIndexFor(ref);
        if (n <= 0) {
            n = 1;
        }
        list_current.setObjectType(&list);
        list_current.setCurrentIndex(n);
    } else {
        list_current.setObjectType(0);     // implies setCurrentIndex=0
    }
    adjustObjectListTop();

    // Update object identifier
    updateObject();
}

/** Update object. Called when object changed. */
void
GChartLocation::updateObject()
{
    GObjectReference obj = getCurrentObject();
    if (obj.isValid()) {
        if (GShip* sh = dynamic_cast<GShip*>(&*obj)) {
            viewport.vec_id = sh->getId();
        } else {
            viewport.vec_id = 0;
        }

        // Update view
        if (dynamic_cast<GShip*>(&*obj)) {
            setViewName("SHIPLOCK");
        } else if (GPlanet* p = dynamic_cast<GPlanet*>(&*obj)) {
            if (p->hasAnyPlanetData())
                setViewName("PLANETLOCK");
            else
                setViewName("UNKNOWNPLANETLOCK");
        } else {
            setViewName(string_t());
        }
    } else {
        setViewName(string_t());
    }
}

inline void
GChartLocation::clearWrapAdjust()
{
    wrap_adjust.x = wrap_adjust.y = 0;
}

/** Adjust list_top member after list/object change. list_top is
    chosen such that the current object is always visible on the list.
    In addition, when there are elements outside the view, at least
    one appears between list_top and the border. */
void
GChartLocation::adjustObjectListTop()
{
    if (list.size() > OBJS_PER_PAGE) {
        int index = list_current.getCurrentIndex();

        /* Check top */
        if (index <= list_top) {
            if (index > 1)
                list_top = index-1;
            else
                list_top = 1;
            changed = true;
        }

        /* Check bottom */
        // index of last item: list_top + getObjectListHeight()-1
        if (index >= list_top + getObjectListHeight() - 1) {
            if (index < int(list.size()))
                list_top = index - getObjectListHeight() + 2;
            else
                list_top = index - getObjectListHeight() + 1;
            changed = true;
        }
    } else {
        /* One page or less: display from first item */
        if (list_top != 1) {
            list_top = 1;
            changed = true;
        }
    }
}

/** Initialize before first load. */
void
GChartLocation::init()
{
    viewport.center.x = viewport.center.y = 2000;
    location.x = location.y = 2000;
    list_current.setObjectType(0);
    list.clear();
    clearWrapAdjust();
    pid = 0;
    changed = true;
}

/** Initialize after game load. */
void
GChartLocation::initAfterLoad()
{
    init();
    updateLocation();
}

/** Initialize upon starchart entry. */
void
GChartLocation::initEntry(const GfxRect& area)
{
    /* initialize viewport */
    viewport.mult = viewport.divi = 1;
    viewport.view = area;
    viewport.vec_id = 0;
    viewport.drawing_filter = viewport.NO_FILTER;
    viewport.init();
    updateObject();
    clearWrapAdjust();

    /* cause initial redraw */
    changed = true;
}

/** Set center of chart. */
void
GChartLocation::setCenter(GPoint pt)
{
    pt = pt.getSimpleCanonicalLocation();

    // Same as last time?
    if (pt == viewport.center)
        return;

    // Change center
    viewport.center = pt;
    changed = true;

    // Update location, it might jump over a map border now
    GPoint near = location.getSimpleNearestAlias(pt);
    if (near != location) {
        wrap_adjust += (near - location);
        location = near;
    }
}

/** Set location. */
void
GChartLocation::setLocation(GPoint pt)
{
    pt = pt.getSimpleCanonicalLocation().getSimpleNearestAlias(viewport.center);

    // Same as last time?
    if (pt == location)
        return;

    // Change location
    location = pt;
    changed = true;
    updateLocation();
}

/** Get cumulative wrap adjustment. If \c location has changed unsuspectedly
    due to a change of the center, the difference will be accumulated here. */
GPoint
GChartLocation::getWrapAdjust() const
{
    return wrap_adjust;
}

/** Get currently-locked object. */
GObjectReference
GChartLocation::getCurrentObject() const
{
    int list_index = list_current.getCurrentIndex();
    if (list_index > 0 && list_index <= int(list.size())) {
        return list[list_index-1]; /* 0-based access! */
    } else {
        return GObjectReference();
    }
}

/** Set currently-locked object. Also adjusts the position if required. */
void
GChartLocation::setCurrentObject(const GObject& obj)
{
    // If it's a map object, go there
    if (const GMapObject* mo = dynamic_cast<const GMapObject*>(&obj)) {
        setLocation(mo->getPos());
    }

    // Focus it.
    int index = list.getIndexFor(obj);
    if (index > 0 && index != list_current.getCurrentIndex()) {
        list_current.setCurrentIndex(index); /* 1-based */
        adjustObjectListTop();
        updateObject();
        changed = true;
    }
}

/** Set name of current view. */
void
GChartLocation::setViewName(string_t name)
{
    if (name != this->view_name) {
        this->view_name = name;
        if (name.size() == 0) {
            // No view means no special keymap
            setKeymap(0);
        } else if (IntKeymapRef km = IntKeymap::getKeymapByName(name)) {
            // There is a keymap matching this view
            setKeymap(km);
        } else {
            // Keep previous keymap
        }
        changed = true;
    }
}

/** Get name of current view. */
string_t
GChartLocation::getViewName() const
{
    return view_name;
}

/** Get Id of currently-locked ship, if any.
    \return ship id, 0 if no ship locked */
int
GChartLocation::getCurrentShipId() const
{
    GObjectReference ref = getCurrentObject();
    if (!ref.isGameObject())
        return 0;

    GShip* sh = dynamic_cast<GShip*>(&*ref);
    if (!sh)
        return 0;

    return sh->getId();
}

/** Perform zooming. */
void
GChartLocation::doZoom(bool enlarge)
{
    if (enlarge)
        changed |= viewport.zoomIn(getDisplayedTurn().getChartBBox());
    else
        changed |= viewport.zoomOut(getDisplayedTurn().getChartBBox());
}

/** Scroll through objects at current location. */
void
GChartLocation::doScroll(int direction, bool mark)
{
    int new_index;
    if (direction > 0) {
        new_index = list.findNextIndexWrap(list_current.getCurrentIndex(), mark);
    } else {
        new_index = list.findPreviousIndexWrap(list_current.getCurrentIndex(), mark);
    }

    if (new_index != 0 && new_index != list_current.getCurrentIndex()) {
        list_current.setCurrentIndex(new_index);
        adjustObjectListTop();
        updateObject();
        changed = true;
    }
}

/** Get possible "other end" position.
    This is a suggested position for the user to go to.
    - if position is at ship, return its waypoint
    - if position is at ship waypoint, return its position
    - if position is in a wormhole, return exit position
    - if circular map is active, switch between map images
    \param [in]  shipId Focus ship Id; can be 0
    \param [out] result Result, alternate position
    \return true on success */
bool
GChartLocation::getOtherPosition(int shipId, GPoint& result) const
{
    GPoint currentPosition = getLocation();
    if (haveDisplayedTurn()) {
        GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
        if (univ.ty_played_ships.isValidIndex(shipId)) {
            const GShip& sh = univ.ty_played_ships.getObjectByIndex(shipId);
            // - if position is at ship, return its waypoint
            GPoint pos = sh.getPos();
            GPoint wp = sh.getWaypoint();
            if (pos == currentPosition && wp != currentPosition) {
                result = wp;
                return true;
            }
            // - if position is at ship waypoint, return its position
            if (wp == currentPosition && pos != currentPosition) {
                result = pos;
                return true;
            }
        }

        // - if position is in a wormhole, return exit position
        // ex chartusr.pas:ChartUfoX
        if (GUfo* ufo = findUfoOtherEnd(currentPosition)) {
            result = ufo->getPos();
            return true;
        }

        // - if circular map is active, switch between map images
        if (chart_conf.getMode() == GChartConfiguration::Circular) {
            GPoint pt = chart_conf.getCanonicalLocation(currentPosition);
            GPoint pt2;
            if (pt != currentPosition) {
                result = pt;
                return true;
            }
            if (chart_conf.getPointAlias(pt, pt2, 1, true)) {
                result = pt2;
                return true;
            }
        }
    }

    // No match
    return false;
}

/** Get height of object list. This limits the object list to a particular size. */
int
GChartLocation::getObjectListHeight() const
{
    if (list.size() > OBJS_PER_PAGE)
        return OBJS_PER_PAGE;
    else
        return list.size();
}

/** Get top index of object list (1-based). */
int
GChartLocation::getObjectListTop() const
{
    return list_top;
}

/** Toggle options. */
void
GChartLocation::toggleOptions(int32 which)
{
    getChartOpts(false, viewport.mult, viewport.divi).toggleOptions(which);
    changed = true;
}

bool
GChartLocation::isChanged() const
{
    return changed;
}

void
GChartLocation::markChanged()
{
    changed = true;
}

void
GChartLocation::clearChanged()
{
    clearWrapAdjust();
    changed = false;
}
