/**
  *  \file client/dialogs/explainpredict.cc
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "client/dialogs/explainpredict.h"
#include "game/ship.h"
#include "game/univ.h"
#include "game/shippredict.h"
#include "game/mission.h"
#include "ui/richdocument.h"
#include "util/unicodechars.h"
#include "ui/layout.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "ui/widgets/button.h"
#include "ui/widgets/keytranslator.h"
#include "game/fcode.h"
#include "cpluslib/format.h"

namespace {
    void startSection(RichDocument& doc, const char* ttl)
    {
        string_t checkMark = UTF_CHECK_MARK " ";
        int checkMarkIndent = fonts[FONT_NORMAL]->getTextWidth(checkMark);

        doc.add(checkMark);
        doc.tabTo(checkMarkIndent);
        doc.setLeftMargin(checkMarkIndent);
        doc.add(ttl);
        doc.addNewline();
    }

    void endSection(RichDocument& doc)
    {
        doc.setLeftMargin(0);
    }
}


void
explainPrediction(GShip& ship, GUniverse& univ)
{
    /* compute stuff */
    GShipTurnPredictor pred(univ, ship.getId());
    pred.addTowee();
    pred.computeMovement();

    GChunnelData chd;
    parseChunnelMission(ship, univ, chd);

    if (pred.getUsedProperties() == 0 && chd.failure == 0) {
        return;
    }

    /* Render text */
    ObjectHolder h;
    UIRichDocument& docWidget(h.add(new UIRichDocument(GfxPoint(100, 100), 0)));
    RichDocument& doc = docWidget.getDocument();

    doc.setPageWidth(ui_screen->getSize().x * 8/10);
    const int mainProps = pred.getUsedProperties();
    if (mainProps != 0) {
        doc.add(RichText(_("Prediction considers...")).withAttribute(RichTextStyleAttribute::Bold));
        doc.addNewline();
        doc.addNewline();

        startSection(doc, string_t(format(_("Movement (%d turn%!1{s%})"), pred.getNumTurns())).c_str());
        endSection(doc);

        // Mission
        if (mainProps & GShipTurnPredictor::UsedMission) {
            startSection(doc, _("Ship mission"));
            if (ship.getMission().isKnown()) {
                const GMission* msn = GMissionList::getInstance().getMissionByShip(ship);
                if (msn) {
                    doc.add(RichText(UIColor::tc_Faded, msn->getLabel(ship)));
                    doc.addNewline();
                }
            }
            endSection(doc);
        }

        // FCode
        if (mainProps & GShipTurnPredictor::UsedFCode) {
            startSection(doc, _("Ship friendly code"));
            GFCodeList& fcl = GFCodeList::getInstance();
            GFCodeList::iterator it = fcl.getFCodeByName(ship.getFCode());
            if (it != fcl.end()) {
                doc.add(RichText(UIColor::tc_Faded, ship.getFCode()).withAttribute(RichTextStyleAttribute::Bold));
                doc.add(RichText(UIColor::tc_Faded, ": " + it->getDescription()));
                doc.addNewline();
            }
            endSection(doc);
        }

        // Alchemy
        if (mainProps & GShipTurnPredictor::UsedAlchemy) {
            startSection(doc, _("Alchemy function"));
            endSection(doc);
        }

        // Shipyard
        if (mainProps & GShipTurnPredictor::UsedShipyard) {
            startSection(doc, _("Starbase shipyard order"));
            endSection(doc);
        }

        // Self repair
        if (mainProps & GShipTurnPredictor::UsedRepair) {
            startSection(doc, _("Supply repair"));
            endSection(doc);
        }

        // Cloak
        if (mainProps & GShipTurnPredictor::UsedCloak) {
            startSection(doc, _("Cloaking"));
            endSection(doc);
        }

        // Damage limit
        if (mainProps & GShipTurnPredictor::UsedDamageLimit) {
            startSection(doc, _("Damage speed limit"));
            endSection(doc);
        }

        // Towee
        if (mainProps & GShipTurnPredictor::UsedTowee) {
            startSection(doc, _("Towed ship's prediction"));
            if (univ.isValidShipId(ship.getTowId())) {
                string_t name = univ.getShip(ship.getTowId()).getName(GObject::PlainName);
                if (!name.empty()) {
                    doc.add(RichText(UIColor::tc_Faded, name));
                    doc.addNewline();
                }
            }
            endSection(doc);
        }
    }

    // Chunnel failures
    if (chd.failure != 0) {
        if (mainProps != 0) {
            doc.addNewline();
        }
        doc.add(RichText(_("Chunnel will fail because...")).withAttribute(RichTextStyleAttribute::Bold));
        doc.addNewline();
        doc.addNewline();
        doc.add(formatChunnelFailures(chd.failure));
    }

    /* Display it */
    docWidget.adjustToDocumentSizeXY();

    UIWindow& window(h.add(new UIWindow(_("Prediction Details"), UIWindow::BLUE_WINDOW, COLOR_GRAY)));
    window.setLayout(UIVBoxLayout::instance5);
    window.add(docWidget);

    UIGroup& g(h.add(new UIGroup(UIHBoxLayout::instance5)));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g.add(h.add(new UISpacer()));
    window.add(g);
    window.add(h.add(new UIKeyTranslator()).add(' ', 1).add(SDLK_ESCAPE, 1));

    window.pack();
    ui_root->centerWidget(window);
    ui_root->runModal(window);
}
