/**
  *  \file client/dialogs/rangeedit.cc
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/rangeedit.h"
#include "ui/window.h"
#include "ui/widgets/decimalsel.h"
#include "ui/layout.h"
#include "client/widgets/optiongrid.h"
#include "cpluslib/holder.h"
#include "ui/widgets/statictext.h"
#include "util/unicodechars.h"
#include "client/widgets/helpwidget.h"
#include "util/nls.h"
#include "ui/uiglobal.h"
#include "gfx/palette.h"
#include "ui/widgets/button.h"
#include "ui/util.h"
#include "util/strtab.h"
#include "game/config.h"
#include "util/strlist.h"
#include "cpluslib/format.h"
#include "game/team.h"
#include "ui/listbox.h"
#include "game/global.h"
#include "game/rangeset.h"
#include "ui/msgbox.h"
#include "game/pref.h"
#include "ui/widgets/quit.h"

namespace {
    ValueEnumParser vep_mode("Own,Ships,Planets,Marked");
    ConfigIntOption opt_chart_range_mode(getUserPreferences(), "Chart.Range.Mode", vep_mode, 0);
    ConfigIntOption opt_chart_range_team(getUserPreferences(), "Chart.Range.Team", ValueBoolParser::instance, 0);
    ConfigIntOption opt_chart_range_distance(getUserPreferences(), "Chart.Range.Distance", ValueIntParser::instance, 0);

    const char* label_table[] = {
        N_("Ranges around"),
        N_("Include team units"),
    };

    const char* value_table[] = {
        N_("no"),
        N_("yes"),
        N_("own units"),
        N_("own ships"),
        N_("own planets"),
        N_("marked units"),
    };

    const int ModeOwn     = 0;
    const int ModeShips   = 1;
    const int ModePlanets = 2;
    const int ModeMarked  = 3;
    const int ModeMax     = 4;

    const int idCancel = 0;
    const int idOK     = 1;
    const int idDropdown = 10;

    const int idRanges = 5;
    const int idTeam   = 6;

    class WRangeDialog : public UIWindow {
     public:
        WRangeDialog();
        ~WRangeDialog();

        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        int getMode() const;
        int getRange() const;
        bool useTeams() const;
        void save() const;

     private:
        ObjectHolder h;
        WOptionGrid* pGrid;
        UIDecimalSelector* pDecimal;

        int mode;
        bool teams;

        void init();
        void update();
        void onClick(int id);
    };

    void addRange(int id, StringList& list, const char* ttl, int value)
    {
        list.add(id, format(_("%s (%d ly)"), ttl, value));
    }
}

WRangeDialog::WRangeDialog()
    : UIWindow(_("Ranges"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      h(),
      pGrid(0),
      pDecimal(0),
      mode(ModeOwn),
      teams(false)
{
    init();
}

WRangeDialog::~WRangeDialog()
{ }

bool
WRangeDialog::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == idDropdown) {
        StringList list;
        int sensorRange    = config.SensorRange(getPlayerId());
        int scanRange      = config.ScanRange(getPlayerId());
        int mineScanRange  = config.MineScanRange(getPlayerId());
        int darkSenseRange = config.DarkSenseRange(getPlayerId());
        addRange(0, list, _("Sensor range"),       sensorRange);
        addRange(1, list, _("Visibility range"),   scanRange);
        addRange(2, list, _("Minefields visible"), mineScanRange);
        addRange(3, list, _("Dark sense"),         darkSenseRange);
        UIStandardListbox box(0,
                              list.getCount(),
                              getTextWidth(*fonts[FONT_NORMAL], list) + 20,
                              list);
        if (box.doStandardMenu(sender)) {
            switch (box.getKey()) {
             case 0: pDecimal->setValue(sensorRange);    mode = ModeShips; break;
             case 1: pDecimal->setValue(scanRange);      mode = ModeOwn;   break;
             case 2: pDecimal->setValue(mineScanRange);  mode = ModeShips; break;
             case 3: pDecimal->setValue(darkSenseRange); mode = ModeShips; break;
            }
            update();
        }
    }
    return UIWindow::handleCommand(cmd, sender);
}

void
WRangeDialog::init()
{
    // VBox
    //   VBox
    //     UIStaticText "Range"
    //     HBox
    //       UIDecimalSelector
    //       UIButton "Down"
    //   WOptionGrid
    //   HBox
    //     "OK"
    //     "Cancel"
    //     UISpacer
    //     "Help"

    setLayout(UIVBoxLayout::instance5);

    // Major widgets
    pGrid = &h.add(new WOptionGrid(getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(label_table, countof(label_table))),
                                   getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(value_table, countof(value_table)))));
    pGrid->addItem(idRanges, 'R', _(label_table[0]));
    pGrid->addItem(idTeam,   'T', _(label_table[1]));
    pDecimal = &h.add(new UIDecimalSelector(0, 0, 1000, 10));

    // Build the dialog
    UIGroup& g1 = h.add(new UIGroup(UIVBoxLayout::instance0));
    g1.add(h.add(new UIStaticText(_("Range [ly]:"), UIStaticText::StaticTitle)));
    UIGroup& g12 = h.add(new UIGroup(UIHBoxLayout::instance0));
    g12.add(*pDecimal);
    g12.add(h.add(new UIButton(idDropdown, UTF_DOWN_ARROW, SDLK_DOWN)).addCommand());
    g1.add(g12);
    add(g1);
    add(*pGrid);
    
    UIGroup& g3 = h.add(new UIGroup(UIHBoxLayout::instance5));
    g3.add(h.add(new UIButton(idOK, _("OK"), SDLK_RETURN)).addStop());
    g3.add(h.add(new UIButton(idCancel, _("Cancel"), SDLK_ESCAPE)).addStop());
    g3.add(h.add(new UISpacer()));
    g3.add(h.add(new UIButton(_("Help"), 'h')));
    add(g3);
    add(h.add(new WHelpWidget("pcc2:starchart:ranges")));
    add(h.add(new UIQuit(idCancel)));
    pack();

    // Load from config
    teams = opt_chart_range_team() != 0;
    mode  = opt_chart_range_mode();
    if (mode < 0 || mode >= ModeMax) {
        mode = 0;
    }
    int dist = opt_chart_range_distance();
    if (dist >= 0 && dist <= 1000) {
        pDecimal->setValue(dist);
    }

    // Connect & initialize
    update();
    pGrid->sig_click.add(this, &WRangeDialog::onClick);
}

void
WRangeDialog::update()
{
    pGrid->find(idRanges).setValue(_(value_table[2 + mode]));
    pGrid->find(idTeam).setValue(_(value_table[teams ? 1 : 0]));
}

void
WRangeDialog::onClick(int id)
{
    switch (id) {
     case idTeam:
        teams = !teams;
        update();
        break;

     case idRanges:
        if (++mode >= ModeMax) {
            mode = 0;
        }
        update();
        break;
    }
}

int
WRangeDialog::getMode() const
{
    return mode;
}

int
WRangeDialog::getRange() const
{
    return pDecimal->getValue();
}

bool
WRangeDialog::useTeams() const
{
    return teams;
}

void
WRangeDialog::save() const
{
    opt_chart_range_team.set(teams);
    opt_chart_range_mode.set(mode);
    opt_chart_range_distance.set(pDecimal->getValue());
}

void
doEditRanges(GRangeSet& ranges)
{
    WRangeDialog dlg;
    ui_root->centerWidget(dlg);
    if (ui_root->runModal(dlg) == idOK) {
        // Save config
        dlg.save();

        // Discard old ranges
        ranges.clear();

        // Initial player set
        GPlayerSet players(getPlayerId());
        if (dlg.useTeams()) {
            for (int i = 1; i <= NUM_OWNERS; ++i) {
                if (getPlayerRelation(i) == is_Ally) {
                    players += i;
                }
            }
        }

        // Build new ranges
        int range = dlg.getRange();
        GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
        switch (dlg.getMode()) {
         case ModeOwn:
            ranges.addObjectType(univ.ty_any_ships, players, false, range);
            ranges.addObjectType(univ.ty_any_planets, players, false, range);
            break;
         case ModeShips:
            ranges.addObjectType(univ.ty_any_ships, players, false, range);
            break;
         case ModePlanets:
            ranges.addObjectType(univ.ty_any_planets, players, false, range);
            break;
         case ModeMarked:
            ranges.addObjectType(univ.ty_any_ships, GPlayerSet::allUpTo(NUM_OWNERS), true, range);
            ranges.addObjectType(univ.ty_any_planets, GPlayerSet::allUpTo(NUM_OWNERS), true, range);
            break;
        }

        // Check emptiness
        if (ranges.isEmpty()) {
            messageBox(_("Your selection does not contain any objects."), _("Ranges"));
        }
    }
}
