/**
  *  \file client/tiles/shiptaskcmd.cc
  *  \brief WShipAutoTaskCommandTile tile
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/shiptaskcmd.h"
#include "client/dialogs/navchart.h"
#include "client/dlg-fcode.h"
#include "client/dlg-mission.h"
#include "client/screens/shiptaskscreen.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/mission.h"
#include "game/racename.h"
#include "ui/richdocument.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "util/misc.h"
#include "util/nls.h"
#include "util/strtab.h"
#include "ui/widgets/decimalsel.h"
#include "ui/commands.h"

static const char*const headings[] = {
    N_("Fuel aboard:"),
    N_("- Movement:"),
    N_("- Cloaking:"),
    N_("= Remaining:"),
};

static const char*const buttons[] = {
    N_("1 - Movement"),
    N_("2 - Cargo"),
    N_("3 - Mission"),
    N_("4 - Misc."),
    N_("O - Options"),
};

WShipAutoTaskCommandTile::WShipAutoTaskCommandTile(GObjectSelection& selection)
    : WTaskMessageTile(GfxPoint(315, 7*font_heights[FONT_NORMAL])),
      atsel(dynamic_cast<WShipAutoTaskSelection*>(&selection)),
      buttonWidth(0)
{
    init();
}

WShipAutoTaskCommandTile::~WShipAutoTaskCommandTile()
{ }

void
WShipAutoTaskCommandTile::init()
{
    // Controls
    setLayout(RightJustified, 0, true);
    StaticTranslatedStringTable tab(buttons, countof(buttons));
    buttonWidth = getTextWidth(*fonts[FONT_SMALL], tab) + 20;
    const int height = font_heights[FONT_SMALL] * 3/2;
    int y = 0;
    for (uint_t i = 0; i < countof(buttons); ++i) {
        UIButton& btn = h.add(new UIButton(1000+i, UIKeyString(tab.get(i))));
        btn.addCommand();
        btn.setState(UIButton::bf_LeftJust + UIButton::bf_SmallSize, true);
        if (i == countof(buttons)-1) {
            y += 5;
        }
        addControl(GfxRect(0, y, buttonWidth, height), btn);
        y += height+2;
    }

    // Init base class (and observer)
    initTask(atsel, buttonWidth);
}

void
WShipAutoTaskCommandTile::drawData(GfxCanvas& can)
{
    // Can we use the base class to draw?
    if (drawTask(can)) {
        return;
    }

    // Sanity checks
    if (atsel == 0) {
        return;
    }
    GShip* sh = dynamic_cast<GShip*>(atsel->getCurrentObject());
    if (sh == 0) {
        return;
    }
    const IntShipPredictor* pred = atsel->getPredictor();

    // Compute layout
    GfxRect area = getInnerSize(5, 5);
    StaticTranslatedStringTable tab(headings, countof(headings));
    int headingWidth = getTextWidth(*fonts[FONT_NORMAL], tab);
    int numberWidth = 6*font_ems[FONT_NORMAL];
    int totalWidth = headingWidth + numberWidth;

    // Make a document
    RichDocument doc;
    doc.setPageWidth(area.w - buttonWidth);

    if (pred != 0) {
        // Regular prediction
        doc.add(tab.get(0));
        doc.addRight(totalWidth, RichText(UIColor::tc_Green, numToString(sh->getCargo(el_Neutronium)) + _(" kt")));
        doc.addNewline();

        doc.add(tab.get(1));
        doc.addRight(totalWidth, RichText(UIColor::tc_Green, numToString(pred->getMovementFuel()) + _(" kt")));
        doc.addNewline();

        doc.add(tab.get(2));
        doc.addRight(totalWidth, RichText(UIColor::tc_Green, numToString(pred->getCloakFuel()) + _(" kt")));
        doc.addNewline();

        doc.add(tab.get(3));
        doc.addRight(totalWidth, RichText(pred->getNumFuelTurns() < pred->getNumTurns() ? UIColor::tc_Red : UIColor::tc_Green,
                                          numToString(pred->getRemainingFuel()) + _(" kt")));
        doc.addParagraph();

        int now = getCurrentTurn().getTurnNumber();
        int age;
        if (pred->getNumFuelTurns() < pred->getNumTurns()) {
            /* getNumFuelTurns() is the number of turns that have fuel.
               Since out-of-fuel happens during the next host run, we
               want it to say "out of fuel next turn" if there is not
               enough fuel for one step. Therefore, we need to add 1.
               This is the same as in PCC 1.x. */
            age = pred->getNumFuelTurns()+1;
            doc.add(RichText(UIColor::tc_Red, format(_("Out of fuel in turn %d"), age + now)));
        } else {
            age = pred->getNumTurns();
            doc.add(RichText(UIColor::tc_Green, format(_("Prediction ends turn %d"), age + now)));
        }
        doc.addNewline();
        if (age == 0) {
            doc.add(_("(this turn)"));
        } else if (age == 1) {
            doc.add(_("(next turn)"));
        } else {
            doc.add(string_t(format(_("(%d turns from now)"), age)));
        }
    } else {
        // leave blank
    }
    doc.finish();

    // Draw it
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    drawBackground(ctx, GfxRect(area.x, area.y, area.w - buttonWidth, area.h));
    doc.draw(ctx, area, 0);
}

void
WShipAutoTaskCommandTile::createSetEnemyCommand()
{
    StringList d;
    d.add(0, _("0 - none"));
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        d.add(i, format("%X - %s", i, player_racenames.getShortName(i)));
    }

    UIStandardListbox lb(0, -1, 240, d);
    if (lb.doStandardListbox(_("Change Primary Enemy"), "pcc2:shiptaskscreen")) {
        atsel->insert(format("SetEnemy %d", lb.getKey()));
    }
}

void
WShipAutoTaskCommandTile::createSetFCodeCommand()
{
    // Find out old friendly code
    IntShipPredictor pred(atsel->getCurrentIndex());
    pred.predictTask(atsel->getEditor(), atsel->getCursor());

    // Dialog
    GFCodeList list(GFCodeList::getInstance(), *atsel->getCurrentObject());
    WFCodeWindow dialog(_("Change Friendly Code"), list);
    dialog.setFCode(pred.getFCode());
    ui_root->centerWidget(dialog);
    if (ui_root->runModal(dialog) == cm_Enter && !dialog.getFCode().empty()) {
        atsel->insert("SetFCode " + quoteString(dialog.getFCode()));
    }
}

void
WShipAutoTaskCommandTile::createSetMissionCommand()
{
    if (GShip* sh = dynamic_cast<GShip*>(atsel->getCurrentObject())) {
        GMissionList msn_list(GMissionList::getInstance(), *sh);
        TMissionData data;
        data.mission = sh->getMission();
        data.args[0] = sh->getInterceptId();
        data.args[1] = sh->getTowId();
        if (chooseMission(msn_list, 0, data)) {
            const char* fmt = (data.args[1]
                               ? "SetMission %d, %d, %d"
                               : data.args[0]
                               ? "SetMission %d, %d"
                               : "SetMission %d");
            atsel->insert(format(fmt, data.mission, data.args[0], data.args[1]));
        }
    }
}

void
WShipAutoTaskCommandTile::createSetSpeedCommand()
{
    // Find current speed
    IntShipPredictor pred(atsel->getCurrentIndex());
    pred.predictTask(atsel->getEditor(), atsel->getCursor());

    // Ask for new speed
    UIDecimalSelector select(0, 0, 9, 1);
    select.setValue(pred.getSpeed());
    if (select.doStandardDialog(_("Set Speed"), _("Warp Factor:"), "pcc2:shiptaskscreen")) {
        atsel->insert(format("SetSpeed %d", select.getValue()));
    }
}

void
WShipAutoTaskCommandTile::createWaypointCommand(const char* cmd, bool setWarp)
{
    // Make sure we have a ship
    GShip* sh = dynamic_cast<GShip*>(atsel->getCurrentObject());
    if (!sh) {
        return;
    }

    // Figure out ship's current state at this place
    IntShipPredictor pred(atsel->getCurrentIndex());
    pred.predictTask(atsel->getEditor(), atsel->getCursor());

    // Ask for new position
    GPoint waypoint = pred.getPosition();
    if (waypoint == sh->getPos()) {
        waypoint = sh->getWaypoint();
    }
    bool hyp = (sh->canDoSpecial(hf_Hyperdrive)
                && pred.getSpeed() > 0
                && pred.getFCode() == "HYP");
    if (!doGenericNavigationChart(getCurrentTurn().getCurrentUniverse(),
                                  pred.getPosition(),
                                  waypoint,
                                  hyp,
                                  square(pred.getSpeed()) * (sh->canDoSpecial(hf_Gravitonic) ? 2 : 1)))
        return;

    atsel->insertWaypointCommand(cmd, waypoint, setWarp, false);
}

void
WShipAutoTaskCommandTile::createCargoTransferCommand(const char* title, const char* verb, const char* all)
{
    static const int8_t types[] = {
        el_Neutronium,
        el_Tritanium,
        el_Duranium,
        el_Molybdenum,
        el_Colonists,
        el_Supplies,
        el_Money,
    };
    static const char letters[] = "NTDMCS$";
    StringList sl;
    for (uint_t i = 0; i < countof(types); ++i) {
        sl.add(i, getCargoName(GCargoType(types[i])));
    }
    sl.add(1000, _("All cargo"));

    UIStandardListbox list(0, sl.getCount(), getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
    if (list.doStandardListbox(title, "pcc2:shiptaskscreen")) {
        if (list.getKey() == 1000) {
            // Unload all cargo
            atsel->insert(format("%s \"10000%s\", \"n\"", verb, all));
        } else {
            // Unload single cargo type
            atsel->insert(format("%s \"10000%c\", \"n\"", verb, letters[list.getKey()]));
        }
    }
}

bool
WShipAutoTaskCommandTile::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    switch (cmd.code) {
     case 1000:
        if (atsel != 0) {
            StringList sl;
            sl.add(1, _("Set Speed"));
            sl.add(2, _("Move To"));
            sl.add(3, _("Move Towards"));
            sl.add(4, _("Set Waypoint"));
            // Generate Route would also be here, but was never implemented in PCC 1.x
            UIStandardListbox list(0, sl.getCount(), getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
            if (list.doStandardMenu(sender)) {
                switch (list.getKey()) {
                 case 1:
                    createSetSpeedCommand();
                    break;
                 case 2:
                    createWaypointCommand("MoveTo", true);
                    break;
                 case 3:
                    createWaypointCommand("MoveTowards", false);
                    break;
                 case 4:
                    createWaypointCommand("SetWaypoint", false);
                    break;
                }
            }
        }
        return true;
     case 1001:
        if (atsel != 0) {
            StringList sl;
            sl.add(1, _("Cargo Unload"));
            sl.add(2, _("Cargo Upload"));
            UIStandardListbox list(0, sl.getCount(), getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
            if (list.doStandardMenu(sender)) {
                switch (list.getKey()) {
                 case 1:
                    createCargoTransferCommand(_("Cargo Unload"), "CargoUnload", "TDMCS$");
                    break;
                 case 2:
                    createCargoTransferCommand(_("Cargo Upload"), "CargoUpload", "NTDMS$");
                    break;
                }
            }
        }
        return true;
     case 1002:
        if (atsel != 0) {
            StringList sl;
            sl.add(1, _("Set Enemy"));
            sl.add(2, _("Set FCode"));
            sl.add(3, _("Set Mission"));
            UIStandardListbox list(0, sl.getCount(), getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
            if (list.doStandardMenu(sender)) {
                switch (list.getKey()) {
                 case 1:
                    createSetEnemyCommand();
                    break;
                 case 2:
                    createSetFCodeCommand();
                    break;
                 case 3:
                    createSetMissionCommand();
                    break;
                }
            }
        }
        return true;
     case 1003:
        if (atsel != 0) {
            atsel->createCommonCommand(sender, true);
        }
        return true;
     case 1004:
        if (atsel != 0) {
            StringList sl;
            if (atsel->isPredictToEnd()) {
                sl.add(1, _("Predict to cursor"));
            } else {
                sl.add(2, _("Predict to end"));
            }
            if (atsel->isShowDistances()) {
                sl.add(3, _("Don't show turn distances"));
            } else {
                sl.add(4, _("Show turn distances"));
            }
            UIStandardListbox list(0, sl.getCount(), getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
            if (list.doStandardMenu(sender)) {
                switch (list.getKey()) {
                 case 1:
                    atsel->setPredictToEnd(false);
                    break;
                 case 2:
                    atsel->setPredictToEnd(true);
                    break;
                 case 3:
                    atsel->setShowDistances(false);
                    break;
                 case 4:
                    atsel->setShowDistances(true);
                    break;
                }
            }
        }
        return true;
    }
    return WTile::handleCommand(cmd, sender);
}

bool
WShipAutoTaskCommandTile::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'g':
            if (atsel != 0) {
                IntShipPredictor pred(atsel->getCurrentIndex());
                pred.predictTask(atsel->getEditor(), atsel->getCursor()+1);
                if (pred.getNumPositions() > 0) {
                    postGoToChart(IntUserInterfaceBinding::get(), pred.getPosition(pred.getNumPositions()-1), 0);
                } else {
                    postGoToChart(IntUserInterfaceBinding::get(), pred.getPosition(), 0);
                }
            }
            return true;
         case 'a':
            if (atsel != 0) {
                createWaypointCommand("MoveTo", true);
            }
            return true;
         case 'e':
            if (atsel != 0) {
                atsel->setPredictToEnd(!atsel->isPredictToEnd());
            }
            return true;
         case 'd':
            if (atsel != 0) {
                atsel->setShowDistances(!atsel->isShowDistances());
            }
            return true;
        }
    }
    return WTile::handleEvent(event, second_pass);
}
