/**
  *  \file client/widgets/expfieldlist.cc
  *  \brief WExportFieldList widget
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <stdexcept>
#include "client/widgets/expfieldlist.h"
#include "client/widgets/expformat.h"
#include "cpluslib/format.h"
#include "game/gamedirs.h"
#include "int/context.h"
#include "int/export.h"
#include "int/export/expconfig.h"
#include "int/propenum.h"
#include "io/dirs.h"
#include "io/glob.h"
#include "io/textfile.h"
#include "ui/dialogs/filedlg.h"
#include "ui/msgbox.h"
#include "ui/util.h"
#include "util/confpars.h"
#include "util/misc.h"
#include "util/nls.h"
#include "util/unicodechars.h"

/** Constructor.
    \param data IntExportFieldList object to edit
    \param ctx Context/array to export. This is used to generate a field list.
               The caller must make sure that this object lives as least as
               long as the WExportFieldList. */
WExportFieldList::WExportFieldList(IntExportFieldList& data, IntContext* ctx)
    : UIListbox(0, 15, 20*font_ems[FONT_NORMAL], data.size()+1),  // xref update
      data(data),
      ctx(ctx),
      format_ctl(0)
{
    sig_dblclick.add(this, &WExportFieldList::editField);
}

/** Update the UIListbox settings after a change. */
void
WExportFieldList::update()
{
    /* "+1" to allow users to scroll past the end and insert there */
    setSize(data.size()+1);
    sig_moved.raise(getId());
    drawWidget();
}

/** User action: change field width.
    \param delta Amount to change */
void
WExportFieldList::changeFieldWidth(int delta)
{
    int i = getCurrent();
    if (i >= 0 && i < int(data.size())) {
        data.changeFieldWidth(i, delta);
        updateCurrent();
    }
}

/** User action: toggle field width. */
void
WExportFieldList::toggleFieldAlignment()
{
    int i = getCurrent();
    if (i >= 0 && i < int(data.size())) {
        data.toggleFieldAlignment(i);
        updateCurrent();
    }
}

/** User action: edit field. */
void
WExportFieldList::editField()
{
    int i = getCurrent();
    if (i >= 0 && i < int(data.size())) {
        string_t fieldName = data.getFieldName(i);
        if (chooseField(fieldName, _("Change field"))) {
            data.setFieldName(i, fieldName);
            updateCurrent();
        }
    } else {
        insertField();
    }
}

/** User action: insert field. */
void
WExportFieldList::insertField()
{
    string_t fieldName;
    int pos = getCurrent();
    if (pos > 0) {
        fieldName = data.getFieldName(pos-1);
    }
    if (chooseField(fieldName, _("Add field"))) {
        // OK
        IntExportFieldList::index_t pos = getCurrent();
        data.add(std::min(pos, data.size()), fieldName, 0);
        update();
        scrollTo(pos+1, 1);
    }
}

bool
WExportFieldList::chooseField(string_t& fieldName, string_t title)
{
    // Build list of possible fields
    class Collector : public IntPropertyAcceptor {
     public:
        StringList names;
        void addProperty(const string_t& name, TIntTypeHint th)
            {
                if (th != thProcedure && th != thFunction && th != thArray) {
                    names.add(0, beautifyName(name));
                }
            }
    };
    Collector coll;
    ctx->enumProperties(coll);
    if (coll.names.getCount() == 0)
        return false;

    // Sort
    coll.names.sortAlphabetically();

    // Determine initial focus
    int index = 0;
    for (int i = 0; i < coll.names.getCount(); ++i) {
        if (strCaseCmp(fieldName, coll.names.get(i)) == 0) {
            index = i;
            break;
        }
    }

    // Ask user
    UIStandardListbox list(0,
                           std::max(3, std::min(15, coll.names.getCount())),
                           getTextWidth(*fonts[FONT_NORMAL], coll.names) + 20,
                           coll.names);
    list.scrollTo(index, 1);
    if (list.doStandardListbox(title, "pcc2:export")) {
        fieldName = coll.names.get(list.getCurrent());
        return true;
    } else {
        return false;
    }
}

/** User action: delete field. */
void
WExportFieldList::deleteField()
{
    IntExportFieldList::index_t pos = getCurrent();
    if (pos < data.size()) {
        data.remove(pos);
        update();
    }
}

/** User action: swap fields.
    \param up true to move current field up, false to move current field down */
void
WExportFieldList::swapFields(bool up)
{
    IntExportFieldList::index_t pos = getCurrent();
    if (up) {
        if (pos > 0) {
            scrollTo(pos-1, 1);
            data.swap(pos, pos-1);
        }
    } else {
        if (pos+1 < data.size()) {
            scrollTo(pos+1, 1);
            data.swap(pos, pos+1);
        }
    }
    drawWidget();
}

void
WExportFieldList::saveSettings()
{
    // FIXME: consider using a different directory (UI.DIRECTORY?)
    Ptr<Directory> dir = game_file_dir.ptr() != 0 ? game_file_dir : Directory::openDir(Directory::getWorkingDirectory());
    UIFileChooserDialog dlg(_("Save Settings"), dir, FileNamePattern::getAllFilesWithExtensionPattern("ccx"));
    ui_root->centerWidget(dlg);
    if (ui_root->runModal(dlg) == 0) {
        return;
    }

    string_t result = appendFileNameExtension(dlg.getResultFile(), "ccx", false);
    try {
        Ptr<Stream> file = Stream::openFile(result, Stream::C_CREATE);
        TextFile tf(*file);
        for (size_t i = 0; i < data.size(); ++i) {
            int fw = data.getFieldWidth(i);
            string_t name = beautifyName(data.getFieldName(i));
            if (fw != 0) {
                tf.writeLine(format("Fields=%s@%d", name, fw));
            } else {
                tf.writeLine(format("Fields=%s", name));
            }
        }
        if (format_ctl != 0) {
            tf.writeLine(format("Charset=%s", format_ctl->getCharsetName()));
            tf.writeLine(format("Format=%s", format_ctl->getTypeCode()));
        }
        tf.sync();
    }
    catch (std::runtime_error& e) {
        messageBox(format(_("Unable to save %s: %s"), result, e.what()), _("Save Settings"));
    }
}

void
WExportFieldList::loadSettings()
{
    // FIXME: consider using a different directory (UI.DIRECTORY?)
    Ptr<Directory> dir = game_file_dir.ptr() != 0 ? game_file_dir : Directory::openDir(Directory::getWorkingDirectory());
    UIFileChooserDialog dlg(_("Load Settings"), dir, FileNamePattern::getAllFilesWithExtensionPattern("ccx"));
    ui_root->centerWidget(dlg);
    if (ui_root->runModal(dlg) == 0) {
        return;
    }

    string_t result = appendFileNameExtension(dlg.getResultFile(), "ccx", false);
    try {
        // Load
        Ptr<Stream> file = Stream::openFile(result, Stream::C_READ);
        IntExportFieldList newFields;
        IntExportConfigParser p(newFields);
        p.parseFile(*file);

        // When we're here, parse succeeded; apply.
        data = newFields;
        if (format_ctl) {
            IntExportOutputType type;
            if (p.getOutputFormat(type)) {
                format_ctl->setOutputType(type);
            }

            int charset_index;
            if (p.getCharsetIndex(charset_index)) {
                format_ctl->setCharsetIndex(charset_index);
            }
        }
        setCurrent(0);
        update();
    }
    catch (std::runtime_error& e) {
        messageBox(format(_("Unable to load %s: %s"), result, e.what()), _("Load Settings"));
    }
}

void
WExportFieldList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);

    int widthWidth = std::min(4*font_ems[FONT_NORMAL], getExtent().w / 2);

    for (int i = from; i < to; ++i) {
        int item = getTop() + i;
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        if (item < int(data.size())) {
            int wi = data.getFieldWidth(item);
            if (wi != 0) {
                outTextF(ctx, area.x + 5, area.y, area.w - widthWidth - 5, beautifyName(data.getFieldName(item)));
                ctx.setTextAlign(2, 0);
                if (wi < 0) {
                    outTextF(ctx, area.x + area.w - 5, area.y, widthWidth - 5, format("%d " UTF_LEFT_ARROW, -wi));
                } else {
                    outTextF(ctx, area.x + area.w - 5, area.y, widthWidth - 5, format(UTF_RIGHT_ARROW " %d", wi));
                }
                ctx.setTextAlign(0, 0);
            } else {
                outTextF(ctx, area.x + 5, area.y, area.w - 5, beautifyName(data.getFieldName(item)));
            }
        } else if (item == int(data.size())) {
            ctx.setColor(UIColor::tc_Faded);
            outTextF(ctx, area.x + 5, area.y, area.w - 5, _("(more...)"));
        } else {
            // ignore
        }
    }
}

bool
WExportFieldList::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case SDLK_DELETE:
            deleteField();
            return true;
         case SDLK_INSERT:
            insertField();
            return true;
         case ' ':
            editField();
            return true;
         case '-' + ss_Shift:
         case SDLK_LEFT + ss_Shift:
            changeFieldWidth(-1);
            return true;
         case '-':
         case SDLK_LEFT:
            changeFieldWidth(e.key.arg ? -e.key.arg : -10);
            return true;
         case '-' + ss_Ctrl:
         case SDLK_LEFT + ss_Ctrl:
            changeFieldWidth(-100);
            return true;
         case '+' + ss_Shift:
         case SDLK_RIGHT + ss_Shift:
            changeFieldWidth(+1);
            return true;
         case '+':
         case SDLK_RIGHT:
            changeFieldWidth(e.key.arg ? e.key.arg : +10);
            return true;
         case '+' + ss_Ctrl:
         case SDLK_RIGHT + ss_Ctrl:
            changeFieldWidth(+100);
            return true;
         case '*':
            toggleFieldAlignment();
            return true;
         case SDLK_UP + ss_Ctrl:
            swapFields(true);
            return true;
         case SDLK_DOWN + ss_Ctrl:
            swapFields(false);
            return true;
         case 's':
         case 's' + ss_Ctrl:
            saveSettings();
            return true;
         case 'r':
         case 'r' + ss_Ctrl:
            loadSettings();
            return true;
        }
    }
    return UIListbox::handleEvent(e, second_pass);
}
