/**
  *  \file game/planetform.cc
  *  \brief Planet Formulas
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/planetform.h"
#include "game/config.h"
#include "game/host.h"
#include "game/formula.h"

/** Compute happiness change target for "Safe Tax" method. */
static int
computeHappinessTarget(int happy)
{
    if (happy >= 100) {
        return 0;
    } else {
        int target = (100 - happy) / 6;
        if (happy < 100 && target == 0) {
            target = 1;
        }
        return target;
    }
}

/*
 *  Structure Formulas
 */

/** Maximum number of structures.
    \param clans      clans on planet
    \param threshold  below threshold, growth is linear; above it's rootic :-) */
static int
getMaxStructuresFormula(int32_t clans, int threshold)
{
    if (clans <= threshold) {
        return clans;
    } else {
        /* the rounding function is actually ERnd, but that never
           makes a difference; sqrt(INT) never ends in .5 exactly */
        return threshold + int(std::sqrt(double(clans - threshold)) + 0.5);
    }
}

/** Get maximum number of structures for a planet.
    This uses the planet's current population. */
mp32_t
getMaxStructures(const GPlanet& pl, TPlanetaryStructure ps)
{
    mn32_t colos = pl.getColonists();
    if (colos.isKnown() && pl.isOwnerKnown()) {
        return mp32_t(getMaxStructures(pl, ps, colos));
    } else {
        return mp32_t();
    }
}

/** Get maximum number of structures for a planet.
    This uses a hyphotetical population. */
int
getMaxStructures(const GPlanet& pl, TPlanetaryStructure ps, int32_t clans)
{
    switch (ps) {
     case str_Factories:
        return getMaxStructuresFormula(clans, 100);
     case str_Mines:
        return getMaxStructuresFormula(clans, 200);
     case str_Defense:
        return getMaxStructuresFormula(clans, 50);
     case str_BaseDefense:
        return pl.hasBase() ? config.MaximumDefenseOnBase(pl.getOwner()) : 0;
    }
    return 0;
}


/*
 *  Colonist Formulas
 */

/** Colonists change.
    \param pl Planet
    \param tax desired tax rate
    \param mifa mines plus factories on planet */
mn32_t
getColonistChange(const GPlanet& pl, int tax, int mifa)
{
    /* can we compute that? */
    if (!pl.getColonists().isKnown() || !pl.isOwnerKnown() || !pl.getTemperature().isKnown())
        return mn32_t();

    double common = 1000 - 80*tax - std::sqrt(double(pl.getColonists()));
    bool   crystal = (raceId(pl.getOwner()) == 7) && config.CrystalsPreferDeserts();
    if (host.isPHost()) {
        if (crystal) {
            return mn32_t(int32_t(common - mifa/3.0 - (100-pl.getTemperature())/0.66) / 100);
        } else {
            return mn32_t(int32_t(common - mifa/3.0 - std::abs(pl.getTemperature()-50)/0.33) / 100);
        }
    } else {
        if (crystal) {
            return mn32_t(int32_t(common - (mifa/3 + 3*(100-pl.getTemperature()))) / 100);
        } else {
            return mn32_t(int32_t(common - (mifa/3 + 3*std::abs(pl.getTemperature()-50))) / 100);
        }
    }
}

/** Colonists happiness change on this planet.
    \return change on this planet in current situation */
mn32_t
getColonistChange(const GPlanet& pl)
{
    if (!pl.getColonistTax().isKnown() || !pl.getStructures(str_Mines).isKnown() || !pl.getStructures(str_Factories).isKnown()) {
        return mn32_t();
    } else {
        return getColonistChange(pl, pl.getColonistTax(), pl.getStructures(str_Mines) + pl.getStructures(str_Factories));
    }
}

/** Get colonist tax amount, as requested. This returns the amount we want
    from the colonists, not the amount we'll finally get.
    \pre pl.isOwnerKnown()
    \pre pl.getColonists().isKnown() */
int32_t
getColonistDue(const GPlanet& pl, int tax)
{
    // Note that these formulas differ in rounding only. PHost uses
    // `Round', THost uses `ERnd' aka `I-don't-care-how-it-rounds'.
    ASSERT(pl.isOwnerKnown() && pl.getColonists().isKnown());
    const int rate = config.ColonistTaxRate(pl.getOwner());
    if (host.isPHost()) {
        return radiv(radiv(pl.getColonists() * tax, 1000) * rate, 100);
    } else {
        return rdivaddl(rdivaddl(pl.getColonists() * tax, 1000, 0) * rate, 100, 0);
    }
}

/** Get colonist tax income.
    \param tax     [in] tax rate
    \param rem_inc [out] this is set to the amount of taxes which we can
    still get from the natives (regarding MaxPlanetaryIncome)
    \return megacredits collected
    \pre pl.isOwnerKnown()
    \pre pl.getColonists().isKnown() */
int32_t
getColonistDueLimited(const GPlanet& pl, int tax, int32_t& rem_inc)
{
    const int32_t due = getColonistDue(pl, tax);
    const int32_t max = config.MaxPlanetaryIncome(pl.getOwner());
    if (due < max) {
        rem_inc = max - due;
        return due;
    } else {
        rem_inc = 0;
        return max;
    }
}

/** Get colonist "safe tax" rate.
    \param mifa Assume this many mines and factories
    \return Safe tax value, or unknown if preconditions not met */
mp16_t
getColonistSafeTax(const GPlanet& pl, int mifa)
{
    // Validate inputs
    if (!pl.isOwnerKnown()
        || !pl.getColonists().isKnown()
        || !pl.getColonistTax().isKnown()
        || !pl.getColonistHappiness().isKnown()
        || !pl.getTemperature().isKnown())
    {
        return mp16_t();
    }

    // Compute result
    int taxlimit = host.getColonistTaxRateLimit(pl.getOwner());
    int tax = taxlimit;
    if (pl.getColonistHappiness() < 70) {
        // Use tax 0 for unhappy colonists
        tax = 0;
    } else {
        // Figure out maximum tax rate yielding a usable happiness:
        int target = computeHappinessTarget(pl.getColonistHappiness());
        while (tax > 0 && getColonistChange(pl, tax, mifa) < target)
            --tax;
    }

    // If higher tax rate produces the same happiness change, use that.
    // This applies when the happiness change goal cannot be reached,
    // so let's use a tax rate that gets some income instead of 0%.
    while (tax < taxlimit && getColonistChange(pl, tax, mifa) == getColonistChange(pl, tax+1, mifa))
        ++tax;

    // If lower tax rates produce the same income, use that:
    int32 limit;
    int32 income = getColonistDueLimited(pl, tax, limit);
    while (tax > 0 && getColonistDueLimited(pl, tax-1, limit) == income)
        --tax;

    return mp16_t(tax);
}

/** Maximum population on planet, for a race.
    \param race player number
    \return maximum population in clans. */
mp32_t
getMaxSupportedColonists(const GPlanet& pl, int player)
{
    bool crystal = (raceId(player) == 7 && config.CrystalsPreferDeserts());
    int32_t limit;
    if (host.isPHost()) {
        if (!config.ClimateLimitsPopulation())
            return mp32_t(250000);
        if (!pl.getTemperature().isKnown())
            return mp32_t();
        if (crystal) {
            if (config.CrystalSinTempBehavior())
                if (pl.getTemperature() >= 15)
                    return mp32_t(int32_t(100000 * std::sin(pl.getTemperature() * PI / 200)));
                else
                    return mp32_t(3 + pl.getTemperature() * config.MaxColTempSlope() / 100);
            else
                return mp32_t(std::max(1, 1000 * pl.getTemperature()));
        }

        if (pl.getTemperature() >= 85)
            limit = config.MaxColTempSlope() * (100-pl.getTemperature()) / 100 + 1;
        else if (pl.getTemperature() <= 14)
            limit = config.MaxColTempSlope() * pl.getTemperature() / 100 + 3;
        else
            limit = int32_t(100000 * std::sin(pl.getTemperature() * PI / 100));
    } else {
        if (!config.ClimateLimitsPopulation())
            return mp32_t(100000);
        if (!pl.getTemperature().isKnown())
            return mp32_t();
        if (crystal)
            return mp32_t(1000 * pl.getTemperature());

        if (pl.getTemperature() >= 85)
            limit = 2 * (100 - pl.getTemperature()) + 1;
        else if (pl.getTemperature() <= 14)
            limit = 2 * pl.getTemperature() + 3;
        else
            limit = int32_t(100000 * std::sin((100 - pl.getTemperature()) * 0.0314) + 0.5);
    }

    if (raceId(player) == 10 && pl.getTemperature() <= 19 && limit < 90000)
        limit = 90000;
    if ((raceId(player) == 4 || raceId(player) >= 9) && pl.getTemperature() >= 84 && limit < 60)
        limit = 60;

    return mp32_t(limit);
}

/** Maximum population on planet.
    \return maximum population in clans. */
mp32_t
getMaxSupportedColonists(const GPlanet& pl)
{
    if (pl.isOwnerKnown()) {
        return getMaxSupportedColonists(pl, pl.getOwner());
    } else {
        return mp32_t();
    }
}

/*
 *  Native Formulas
 */

static inline double
pdiv(int a, int b)
{
    if (host.isPHost()) {
        return double(a) / b;
    } else {
        return a / b;
    }
}

/** Native happiness change.
    \returns happiness change for current situation. */
mn32_t
getNativeChange(const GPlanet& pl)
{
    if (!pl.getNativeTax().isKnown() || !pl.getStructures(str_Mines).isKnown() || !pl.getStructures(str_Factories).isKnown()) {
        return mn32_t();
    } else {
        return getNativeChange(pl, pl.getNativeTax(), pl.getStructures(str_Mines) + pl.getStructures(str_Factories));
    }
}

/** Native happiness change.
    \param tax   desired tax rate
    \param mifa  mines plus factories on planet */
mn32_t
getNativeChange(const GPlanet& pl, int tax, int mifa)
{
    /* formula is correct for PHost and THost */
    if (!pl.getNativeGovernment().isKnown() || !pl.getNatives().isKnown())
        return mn32_t();

    int change =
        int32(500 + 50 * pl.getNativeGovernment()
              - 85 * tax
              - pdiv(mifa, 2)
              - std::sqrt(double(pl.getNatives())))
        / 100;

    if (pl.getNativeRace().isKnown() && pl.getNativeRace() == nr_Avian) {
        change += 10;
    }
    return mn32_t(change);
}

/** Get native tax amount, as requested, for this planet. This returns
    the amount we're asking from them, not what we'll get.
    \pre native race, government, population, and planet owner known */
int32
getNativeDue(const GPlanet& pl, int tax)
{
    ASSERT(pl.getNativeRace().isKnown() && pl.getNativeGovernment().isKnown());
    ASSERT(pl.getNatives().isKnown() && pl.isOwnerKnown());

    return getNativeDue(tax, pl.getNativeRace(), pl.getNativeGovernment(), pl.getNatives(), pl.getOwner());
}

/** Get native tax amount, as requested, parameterized. This returns
    the amount we're asking from them, not what we'll get.
    \param tax    tax rate
    \param race   native race
    \param gov    government factor (SPI)
    \param pop    native clans
    \param owner  owner (for production rates) */
int32_t
getNativeDue(int tax, int race, int gov, int32_t pop, int owner)
{
    int32_t due;
    if (host.isPHost()) {
        due = radiv(radiv(tax * gov * pop, 5000) * config.NativeTaxRate(owner), 100);
    } else {
        due = rdivaddl(rdivaddl(tax * gov * pop, 5000, 0) * config.ColonistTaxRate(owner), 100, 0);
    }
    if (race == nr_Insectoid) {
        return 2*due;
    } else {
        return due;
    }
}

/** Get native tax amount, limited. Limits the tax income by income
    limit and available colonists.
    \param tax     tax rate
    \param rem_inc remaining allowed income; output of getColonistDueLimited(). */
int32_t
getNativeDueLimited(const GPlanet& pl, int tax, int32_t rem_inc)
{
    /* amorphs don't pay */
    if (pl.getNativeRace() == nr_Amorphous)
        return 0;

    /* cyborgs can only tax to 20 */
    int limit = host.getNativeTaxRateLimit(pl.getOwner());
    if (tax > limit)
        tax = limit;

    /* normal formulas here */
    int32_t due = getNativeDue(pl, tax);
    int32_t colos = pl.getColonists();

    if (host.isPHost()) {
        if (pl.getNativeRace() == nr_Insectoid)
            colos *= 2;
        colos = radiv(colos * config.NativeTaxRate(pl.getOwner()), 100);
    } else {
        colos = colos * config.ColonistTaxRate(pl.getOwner()) / 100;
        if (pl.getNativeRace() == nr_Insectoid)
            colos *= 2;
    }

    if (due > colos)
        due = colos;
    if (due > rem_inc)
        due = rem_inc;
    return due;
}

/** Get native "safe tax" rate.
    \param mifa Assume this many mines and factories
    \return Safe tax value, or unknown if preconditions not met */
mp16_t
getNativeSafeTax(const GPlanet& pl,int mifa)
{
    // Validate inputs
    if (!pl.isOwnerKnown()
        || !pl.getColonists().isKnown()
        || !pl.getNativeRace().isKnown()
        || !pl.getNatives().isKnown()
        || !pl.getNativeGovernment().isKnown()
        || !pl.getNativeHappiness().isKnown()
        || pl.getNatives() <= 0)
    {
        return mp16_t();
    }

    // Compute result
    int taxlimit = host.getNativeTaxRateLimit(pl.getOwner());
    int tax = taxlimit;
    if (pl.getNativeHappiness() < 70 || pl.getNativeRace() == nr_Amorphous) {
        // Use tax 0 for unhappy natives or Amorphs
        tax = 0;
    } else {
        // Figure out maximum tax rate yielding a usable happiness:
        int target = computeHappinessTarget(pl.getNativeHappiness());
        while (tax > 0 && getNativeChange(pl, tax, mifa) < target)
            --tax;
    }

    // If higher tax rate produces the same happiness change, use that.
    // This applies when the happiness change goal cannot be reached,
    // so let's use a tax rate that gets some income instead of 0%.
    while (tax < taxlimit && getNativeChange(pl, tax, mifa) == getNativeChange(pl, tax+1, mifa))
        ++tax;

    // If lower tax rates produce the same income, use that.
    // Note that the original code tried (and failed) to handle the relation
    // between colonist tax, native tax, and MaxPlanetaryIncome here. Doing
    // this correctly means handling the (assumed) colonist tax rate here,
    // which would complicate matters too much for my taste. This code matters
    // in two places: where safe-tax income hits MaxPlanetaryIncome (rare),
    // and where population is low enough to make rounding effects matter.
    int32_t limit = 0x7FFFFFFF;
    int32_t income = getNativeDueLimited(pl, tax, limit);
    while (tax > 0 && getNativeDueLimited(pl, tax-1, limit) == income)
        --tax;

    return mp16_t(tax);
}

mp16_t
getNativeBaseTax(const GPlanet& pl, int happyTarget)
{
    if (!pl.isOwnerKnown()) {
        return mp16_t();
    } else {
        return getNativeBaseTax(pl, pl.getOwner(), happyTarget);
    }
}

/** Get native "base tax" rate. The base tax assumes a fixed mines/factories
    count to make results comparable between planets.
    \param happyTarget Happiness target (0=base tax, -30=max tax)
    \return tax rate, unknown if preconditions not met */
mp16_t
getNativeBaseTax(const GPlanet& pl, int player, int happyTarget)
{
    // Check preconditions
    if (!pl.getNativeRace().isKnown()
        || !pl.getNatives().isKnown()
        || !pl.getNativeGovernment().isKnown())
    {
        return mp16_t();
    }

    // Acquire data
    int nid = pl.getNativeRace();
    int32_t natives = pl.getNatives();
    int spi = pl.getNativeGovernment();

    // No taxes if natives wouldn't pay anyway, or there are none
    if (natives == 0 || nid == nr_Amorphous) {
        return mp16_t(0);
    }

    // Compute result
    const int mifa = 100;
    int tax = host.getNativeTaxRateLimit(player);
    while (tax > 0) {
        /* Host 3.22.040 formula:
           Trunc(1000.0 - sqrt(Planet.Natives) - 85 * Planet.NativeTax
                        - (Planet.Mines + .Factories) DIV 2
                        - 50*(10-Planet.NativeGov))
              DIV 100
           change:=5 + Trunc(spi/2 - (Sqrt(Natives) + 85*LONGINT(tax) + 100/2) / 100);
           change:=5 + Trunc(50*spi - (85*LONGINT(tax) + 50) - Sqrt(Natives)) DIV 100;
           since MIFA is even, the same formula can be used for PHost and THost. */
        int change = int32_t(500 - (mifa/2) + (50*spi) - (85*tax) - std::sqrt(double(natives))) / 100;
        if (nid == nr_Avian)
            change += 10;
        if (change >= happyTarget)
            break;
        --tax;
    }
    return mp16_t(tax);
}

/** Get Bovinoid supply contribution. Contribution is not yet limited by
    colonists. */
int32_t
getBovinoidSupplyContribution(const GPlanet& pl)
{
    if (pl.getNativeRace().isKnown() && pl.isOwnerKnown() && pl.getNatives().isKnown())
        if (pl.getNativeRace() == nr_Bovinoid)
            return getBovinoidSupplyContribution(pl.getNatives(), pl.getOwner());
        else
            return 0;
    else
        return 0;
}

/** Get Bovinoid supply contribution, parameterized. */
int32_t
getBovinoidSupplyContribution(int32_t pop, int owner)
{
    // FIXME: for ultimate accuracy, we have to consider factories as
    // well. PHost does "(bovi + factories) * ProductionRate", not
    // "bovi * ProductionRate + factories * ProductionRate".
    if (host.isPHost()) {
        return (pop / 100) * config.ProductionRate(owner) / 100;
    } else {
        return pop / 100;
    }
}

/** Get Bovinoid supply contribution. This returns the amount of
    supplies finally collected. */
int32_t
getBovinoidSupplyContributionLimited(const GPlanet& pl)
{
    // FIXME: the same problem as in getBovinoidSupplyContribution(int,int)
    // appears here. We *could* add a workaround here (compute totals,
    // compute factory contribution, compute difference), but is it
    // really worth to do that for an accuracy of +/-1 with a config
    // option no-one uses?
    //      -- update: it's going to be used in PList 2.5, so someday
    //         it should be supported
    if (!pl.getNativeRace().isKnown() || !pl.getColonists().isKnown())
        return 0;

    if (pl.getNativeRace() == nr_Bovinoid) {
        int32_t limit = host.isPHost() ? pl.getColonists() * config.ProductionRate(pl.getOwner()) / 100 : pl.getColonists();
        int32_t due = getBovinoidSupplyContribution(pl);
        if (due < limit) {
            return due;
        } else {
            return limit;
        }
    } else {
        return 0;
    }
}

/** Get Amorphous colonist breakfast. Returns the number of clans
    eaten by our lovely amorphous natives when they have a happiness
    as specified. */
int32_t
getAmorphousBreakfast(int happy)
{
    if (host.isPHost()) {
        if (happy >= 70)
            return 5;
        else if (happy >= 50)
            return 20;
        else
            return 40;
    } else {
        if (happy > 90)
            return 5;
        else
            return 95 - happy;
    }
}


/*
 *  Mining Formulas
 */

/** Get mining capacity.
    \param index Mineral index
    \param mines Number of mines on planet
    \pre density is known */
int
getMiningCapacity(const GPlanet& pl, int index, int mines)
{
    // Gather inputs
    int density = pl.getOreDensity(index);

    int mining_rate;
    if (pl.isOwnerKnown()) {
        mining_rate = config.RaceMiningRate(pl.getOwner());
    } else {
        mining_rate = 100;
    }

    int reptile_factor;
    if (pl.getNativeRace().isKnown()
        && pl.getNativeRace() == nr_Reptilian
        && (!pl.getNatives().isKnown() || pl.getNatives() > 0))
    {
        reptile_factor = 2;
    } else {
        reptile_factor = 1;
    }

    // Host-dependant formula
    if (!host.isPHost()) {
        return rdivaddl(rdivaddl(density * int32_t(mines), 100, 0) * mining_rate, 100, 0) * reptile_factor;
    } else {
        int add = (host.isPHostRoundingMiningResults() ? 50 : 0);
        return (((density * int32_t(mining_rate) + add) / 100) * reptile_factor * mines + add) / 100;
    }
}
