/**
  *  \file game/save-cksum.cc
  *  \brief Game Saving: Checksums
  *
  *  This module contains routines to compute checksums needed to
  *  create valid game directories. These routines are similar to
  *  those in game/save.cc, but do not produce output on a stream.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/save.h"
#include "game/checksum.h"
#include "game/struct.h"
#include "game/univ.h"
#include "game/ship.h"
#include "game/planet.h"

/** Compute checksum for all ships. Selects a subset of all ships stored
    in \c univ by masking their source flags with \c src, and computes
    the standard checksum over their binary representation.
    \param univ Contains the ships
    \param src Selection
    \param flags save_ChecksumCount: include 16-bit object count in checksum
    \return checksum */
uint32_t
computeShipChecksum(const GUniverse& univ, const GPlayerSet src, const int flags)
{
    uint32_t checksum = 0;
    uint16_t count = 0;
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        const GShip& s = univ.getShip(i);
        if (s.getShipSource().containsAnyOf(src)) {
            char buffer[TShip::size];
            TShip tmp;
            ++count;
            s.getShipData(tmp);
            storeStructure(buffer, tmp);
            checksum += getChecksum(buffer, sizeof buffer);
        }
    }
    if (flags & save_ChecksumCount)
        checksum += getChecksum(count);
    return checksum;
}

/** Compute checksum for all planets.
    \param univ Contains the planets
    \param src Selection
    \param flags save_ChecksumCount: include 16-bit object count in checksum
    \see computeShipChecksum()
    \return checksum */
uint32_t
computePlanetChecksum(const GUniverse& univ, const GPlayerSet src, const int flags)
{
    uint32_t checksum = 0;
    uint16_t count = 0;
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        const GPlanet& p = univ.getPlanet(i);
        if (p.getPlanetSource().containsAnyOf(src)) {
            char buffer[TPlanet::size];
            TPlanet tmp;
            ++count;
            p.getPlanetData(tmp);
            storeStructure(buffer, tmp);
            checksum += getChecksum(buffer, sizeof buffer);
        }
    }
    if (flags & save_ChecksumCount)
        checksum += getChecksum(count);
    return checksum;
}

/** Compute checksum for all starbases.
    \param univ Contains the starbases
    \param src Selection
    \param flags save_ChecksumCount: include 16-bit object count in checksum
    \see computeShipChecksum()
    \return checksum */
uint32_t
computeBaseChecksum(const GUniverse& univ, const GPlayerSet src, const int flags)
{
    uint32_t checksum = 0;
    uint16_t count = 0;
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        const GPlanet& p = univ.getPlanet(i);
        if (p.getBaseSource().containsAnyOf(src)) {
            char buffer[TStarbase::size];
            TStarbase tmp;
            ++count;
            p.getBaseData(tmp);
            storeStructure(buffer, tmp);
            checksum += getChecksum(buffer, sizeof buffer);
        }
    }
    if (flags & save_ChecksumCount)
        checksum += getChecksum(count);
    return checksum;
}
