/**
  *  \file int/export/exptext.cc
  *  \brief Export to Text File
  *
  *  This implements the 'text' and 'table' output formats.
  *  It accumulates a complete line at a time, then trims
  *  whitespace, and outputs the result.
  *
  *  Modelled partially after PCC 1.x export.pas::CTextExporter
  *  (PCC 1.x doesn't have 'text', only 'table').
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include "int/export/exptext.h"
#include "cpluslib/utf8.h"
#include "int/value.h"
#include "io/textfile.h"

/** Get default width of a particular column type. */
static int
getDefaultWidth(TIntTypeHint th)
{
    switch (th) {
     case thBool:
        return 3;
     case thInt:
        return 10;
     case thFloat:
        return 10;
     case thString:
        return 30;
     case thNone:
        return 100;
     default:
        return 30;
    }
}

/** Constructor.
    \param tf Text file to write to
    \param boxes true to create boxy table, false to create simple table */
IntTextExporter::IntTextExporter(TextFile& tf, bool boxes)
    : tf(tf),
      boxes(boxes),
      line(),
      widths(),
      totalWidth(0)
{ }

/** Start a new data line. */
void
IntTextExporter::startLine()
{
    line.clear();
    currentField = 0;
    if (boxes) {
        line.append("| ");
    }
}

/** Add a value.
    \param value Formatted value
    \param left true to left-justify, false to right-justify. */
void
IntTextExporter::addValue(string_t value, bool left)
{
    // Remove newlines
    string_t::size_type nn = value.find('\n');
    if (nn != value.npos) {
        value.erase(nn);
        value.append("...");
    }

    // Update alignment from configuration
    if (int align = alignments[currentField]) {
        left = (align < 0);
    }

    // Format into field
    std::size_t n = lengthUtf8(value);
    std::size_t width = widths[currentField];
    if (n != width) {
        if (n < width) {
            if (!left) {
                line.append(width - n, ' ');
            }
            line.append(value);
            if (left) {
                line.append(width - n, ' ');
            }
        } else {
            if (left) {
                line.append(substrUtf8(value, 0, width));
            } else {
                line.append(substrUtf8(value, n - width, width));
            }
        }
    } else {
        line.append(value);
    }

    if (boxes) {
        line.append(" | ");
    } else {
        line.append(" ");
    }

    currentField++;
}

/** End a line. This outputs the line. */
void
IntTextExporter::endLine()
{
    tf.writeLine(strRTrim(line));
}

/** Write a divider line. */
void
IntTextExporter::writeDivider()
{
    std::size_t total = totalWidth;
    if (boxes) {
        // 2 spaces in each field, plus one '|' per field, plus one extra '|'
        total += 3*widths.size() + 1;
    } else {
        // 1 extra space per field, except for last one
        total += widths.size() - 1;
    }
    tf.writeLine(string_t(total, '-'));
}

void
IntTextExporter::startTable(const IntExportFieldList& fields, const TIntTypeHint types[])
{
    // Populate widths
    totalWidth = 0;
    for (IntExportFieldList::index_t i = 0; i < fields.size(); ++i) {
        std::size_t ele;
        if (fields.getFieldWidth(i) != 0) {
            ele = std::abs(fields.getFieldWidth(i));
        } else {
            ele = getDefaultWidth(types[i]);
        }
        totalWidth += ele;
        widths.push_back(ele);
        alignments.push_back(fields.getFieldWidth(i));
    }

    // Write headings
    startLine();
    for (IntExportFieldList::index_t i = 0; i < fields.size(); ++i) {
        addValue(fields.getFieldName(i), types[i] != thInt && types[i] != thFloat);
    }
    endLine();
    currentLine = 0;
}

void
IntTextExporter::startRecord()
{
    // Divider
    if (currentLine == 0 || (boxes && (currentLine % 10 == 0))) {
        writeDivider();
    }
    ++currentLine;

    startLine();
}

void
IntTextExporter::addField(IntValue* value, const string_t& /*name*/, TIntTypeHint /*type*/)
{
    IntIntValue* iv = dynamic_cast<IntIntValue*>(value);
    IntFloatValue* fv = dynamic_cast<IntFloatValue*>(value);
    bool left = ((iv == 0 || iv->isBool()) && fv == 0);
    addValue(IntValue::toStringOf(value, false), left);
}

void
IntTextExporter::endRecord()
{
    endLine();
}

void
IntTextExporter::endTable()
{
    if (boxes) {
        writeDivider();
    }
}
