/**
  *  \file int/opcode.cc
  *  \brief Interpreter: Opcodes
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/opcode.h"
#include "cpluslib/misc.h"
#include "int/unary.h"
#include "int/binary.h"
#include "int/ternary.h"

static const char*const scopes[]     = { "var", "loc", "top", "glob", "gvar", "lit", "int", "bool" };
static const char*const scope_args[] = { "%n",  "%L",  "%T",  "%G",   "%n",   "%l",  "%d",  "%d" };
static const char*const ims[]        = { "call",  "load",  "store", "pop",          // Refuse nothing --> call anything
                                         "proc",  "pload", "pstore", "ppop",        // RefuseFunctions --> procedures
                                         "fcall", "func",  "fstore", "fpop" };      // RefuseProcedures --> functions
static const char*const stack[]      = { "dup", "drop", "swap" };

static const char*const specials[] = {
    "uncatch",
    "return\t%u",
    "with",
    "endwith",
    "firstindex",
    "nextindex",
    "endindex",
    "evals\t%u",
    "evalx",
    "defsub\t%n",
    "defshipp\t%n",
    "defplanetp\t%n",
    "load",
    "print",
    "addhook",
    "runhook",
    "throw",
    "terminate",
    "suspend",
    "newarray\t%u",
    "makelist\t%u",
    "newhash",
    "instance",
    "resizearray\t%u",
    "bind",
    "first",
    "next",
};

static const char*
formatEnum(uint8_t minor, const char*const* values, std::size_t count)
{
    return minor < count ? values[minor] : "?";
}

static inline const char*
formatScope(uint8_t minor)
{
    return formatEnum(minor, scopes, countof(scopes));
}

static inline const char*
formatIM(uint8_t minor)
{
    return formatEnum(minor, ims, countof(ims));
}

/** Get template for disassembling this opcode. The caller must fill
    in the placeholders from the opcode's \c arg field.
    - %n name table index
    - %l literal table index
    - %s subroutine table index
    - %u unsigned int
    - %d signed int
    - %L local variable name, given by address
    - %T static variable name, given by address
    - %G shared variable name, given by address */
void
IntOpcode::getDisassemblyTemplate(string_t& tpl) const
{
    switch (major) {
     case maPush:
        tpl += "push";
        tpl += formatScope(minor);
        tpl += "\t";
        tpl += formatEnum(minor, scope_args, countof(scope_args));
        break;
     case maBinary:
        tpl += "b";
        tpl += getBinaryName(minor);
        break;
     case maUnary:
        tpl += "u";
        tpl += getUnaryName(minor);
        break;
     case maTernary:
        tpl += "t";
        tpl += getTernaryName(minor);
        break;
     case maJump:
        {
            uint8_t flags = minor & ~jSymbolic;
            if (flags == 0) {
                tpl += "label";
            } else if (flags < jCatch) {
                tpl += "j";
                if ((flags & jAlways) != jAlways) {
                    if (flags & jIfTrue)
                        tpl += "t";
                    if (flags & jIfFalse)
                        tpl += "f";
                    if (flags & jIfEmpty)
                        tpl += "e";
                    if ((flags & jAlways) == 0)
                        tpl += "never"; /* might occur as jneverp */
                }
                if (flags & jPopAlways)
                    tpl += "p";
            } else if (flags == jCatch) {
                tpl += "catch";
            } else if (flags == jDecZero) {
                tpl += "jdz";
            } else {
                tpl += "junknown";
            }

            if (minor & jSymbolic)
                tpl += "\tsym%u";
            else
                tpl += "\t#%u";
        }
        break;
     case maIndirect:
        tpl += formatIM(minor);
        tpl += "ind\t%u";
        break;
     case maStack:
        tpl += formatEnum(minor, stack, countof(stack));
        tpl += "\t%u";
        break;
     case maPop:
        tpl += "pop";
        tpl += formatScope(minor);
        tpl += "\t";
        tpl += formatEnum(minor, scope_args, countof(scope_args));
        break;
     case maStore:
        tpl += "store";
        tpl += formatScope(minor);
        tpl += "\t";
        tpl += formatEnum(minor, scope_args, countof(scope_args));
        break;
     case maMemref:
        tpl += formatIM(minor);
        tpl += "mem\t%n";
        break;
     case maDim:
        tpl += "dim";
        tpl += formatScope(minor);
        tpl += "\t%n";
        break;
     case maSpecial:
        tpl += "s";
        tpl += formatEnum(minor, specials, countof(specials));
        break;

     case maFusedUnary:
        tpl += "push";
        tpl += formatScope(minor);
        tpl += "(u)\t";
        tpl += formatEnum(minor, scope_args, countof(scope_args));
        break;

     case maFusedBinary:
        tpl += "push";
        tpl += formatScope(minor);
        tpl += "(b)\t";
        tpl += formatEnum(minor, scope_args, countof(scope_args));
        break;

     case maFusedComparison:
        tpl += "b";
        tpl += getBinaryName(minor);
        tpl += "(j)";
        break;

     case maFusedComparison2:
        tpl += "push";
        tpl += formatScope(minor);
        tpl += "(b,j)\t";
        tpl += formatEnum(minor, scope_args, countof(scope_args));
        break;

     case maInplaceUnary:
        tpl += "push";
        tpl += formatScope(minor);
        tpl += "(xu)\t";
        tpl += formatEnum(minor, scope_args, countof(scope_args));
        break;

     default:
        tpl += "unknown\t%u";
        break;
    }
}
