/**
  *  \file u/t_arch_file.cc
  *  \brief Testsuite for architecture-dependant file access functions
  *
  *  Note that this assumes that the current directory is writable.
  *
  *  After checking the size of a file, we also check its position again.
  *  This is to catch errors in implementations of getSize() that seek to
  *  the file end to figure out the size, to make sure they rewind the file
  *  correctly.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "io/streams.h"     /* this declares the create function */

#include <cstdio>
#include <cstring>
#include "u/t_arch.h"

static const char FILE_NAME[] = "00test00.dat";

ArchFileTestSuite::ArchFileTestSuite()
{ }

/** Destructor. Makes sure that the file no longer exists. */
ArchFileTestSuite::~ArchFileTestSuite()
{
    std::remove(FILE_NAME);
}

void
ArchFileTestSuite::doCreateFile()
{
    // Open the file
    Ptr<Stream> f = Stream::openFile(FILE_NAME, Stream::C_CREATE | Stream::C_WRITE);
    TS_ASSERT(f != 0);
    TS_ASSERT_EQUALS(f->getPos(), 0);
    TS_ASSERT_EQUALS(f->getSize(), 0);
    TS_ASSERT_EQUALS(f->getPos(), 0);
    TS_ASSERT_EQUALS(int(f->getCapabilities() & Stream::C_WRITE), int(Stream::C_WRITE));
    TS_ASSERT(f->canDo(Stream::C_WRITE));

    // Write 20 bytes
    int bytesWritten = f->write("abcdefghijklmnopqrst", 20);
    TS_ASSERT_EQUALS(bytesWritten, 20);
    TS_ASSERT_EQUALS(f->getPos(), 20);
    TS_ASSERT_EQUALS(f->getSize(), 20);
    TS_ASSERT_EQUALS(f->getPos(), 20);

    // Test seeking
    TS_ASSERT_THROWS_NOTHING(f->seek(10));
    TS_ASSERT_EQUALS(f->getPos(), 10);
    TS_ASSERT_EQUALS(f->getSize(), 20);
    TS_ASSERT_EQUALS(f->getPos(), 10);

    // Write 20 more bytes
    bytesWritten = f->write("ABCDEFGHIJKLMNOPQRST", 20);
    TS_ASSERT_EQUALS(bytesWritten, 20);
    TS_ASSERT_EQUALS(f->getPos(), 30);
    TS_ASSERT_EQUALS(f->getSize(), 30);
    TS_ASSERT_EQUALS(f->getPos(), 30);

    // Seek again
    TS_ASSERT_THROWS_NOTHING(f->seek(5));
    TS_ASSERT_EQUALS(f->getPos(), 5);
    TS_ASSERT_EQUALS(f->getSize(), 30);
    TS_ASSERT_EQUALS(f->getPos(), 5);

    // And write again
    bytesWritten = f->write("01234", 5);
    TS_ASSERT_EQUALS(bytesWritten, 5);
    TS_ASSERT_EQUALS(f->getPos(), 10);
    TS_ASSERT_EQUALS(f->getSize(), 30);
    TS_ASSERT_EQUALS(f->getPos(), 10);
}

void
ArchFileTestSuite::doReadFile(const char content[])
{
    using std::strlen;

    // Open the file
    Ptr<Stream> f = Stream::openFile(FILE_NAME, Stream::C_READ);
    TS_ASSERT(f != 0);
    TS_ASSERT_EQUALS(f->getPos(), 0);
    TS_ASSERT_EQUALS(f->getSize(), (long)strlen(content));
    TS_ASSERT_EQUALS(f->getPos(), 0);
    TS_ASSERT(f->canDo(Stream::C_READ));

    // Read character by character
    char c;
    int n = 0;
    for (const char* p = content; *p != 0; ++p) {
        int bytesRead = f->read(&c, 1);
        TS_ASSERT_EQUALS(bytesRead, 1);
        TS_ASSERT_EQUALS(c, *p);
        ++n;
        TS_ASSERT_EQUALS(f->getPos(), n);
    }
    n = f->read(&c, 1);
    TS_ASSERT_EQUALS(n, 0);

    // Try the same after seeking
    n = 12;
    TS_ASSERT_THROWS_NOTHING(f->seek(n));
    for (const char* p = content + n; *p != 0; ++p) {
        int bytesRead = f->read(&c, 1);
        TS_ASSERT_EQUALS(bytesRead, 1);
        TS_ASSERT_EQUALS(c, *p);
        ++n;
        TS_ASSERT_EQUALS(f->getPos(), n);
    }
    n = f->read(&c, 1);
    TS_ASSERT_EQUALS(n, 0);
}

void
ArchFileTestSuite::doModifyFile()
{
    // Open the file
    Ptr<Stream> f = Stream::openFile(FILE_NAME, Stream::C_WRITE);
    TS_ASSERT(f != 0);
    TS_ASSERT_EQUALS(f->getPos(), 0);
    TS_ASSERT_EQUALS(f->getSize(), 30);
    TS_ASSERT_EQUALS(f->getPos(), 0);
    TS_ASSERT(f->canDo(Stream::C_WRITE));

    // Write something at the beginning
    int bytesWritten = f->write("0123456789", 10);
    TS_ASSERT_EQUALS(bytesWritten, 10);
    TS_ASSERT_EQUALS(f->getPos(), 10);
    TS_ASSERT_EQUALS(f->getSize(), 30);
    TS_ASSERT_EQUALS(f->getPos(), 10);

    // Write something in the middle
    TS_ASSERT_THROWS_NOTHING(f->seek(15));
    bytesWritten = f->write("ZZZ", 3);
    TS_ASSERT_EQUALS(bytesWritten, 3);
    TS_ASSERT_EQUALS(f->getPos(), 18);
    TS_ASSERT_EQUALS(f->getSize(), 30);
    TS_ASSERT_EQUALS(f->getPos(), 18);
}

void
ArchFileTestSuite::testFileAccess()
{
    doCreateFile();
    doReadFile("abcde01234ABCDEFGHIJKLMNOPQRST");
    doModifyFile();
    doReadFile("0123456789ABCDEZZZIJKLMNOPQRST");
}
