/**
  *  \file util/strlist.cc
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "util/strlist.h"

/******************************* StringList ******************************/

/** \class StringList
    \brief Ordered key/string pair list

    This provides a container of pairs of keys (integers) and strings,
    to be used for list boxes and similar things. */

StringList::StringList()
{ }

StringList::~StringList()
{ }

/** Add a key/string pair. */
void
StringList::add(int key, const string_t& s)
{
    data.push_back(std::make_pair(key, s));
}

/** Add a key/string pair. */
void
StringList::add(int key, const char* s)
{
    add(key, string_t(s));
}

/** Add a key/string pair at front. */
void
StringList::addAtFront(int key, const string_t& s)
{
    data.push_front(std::make_pair(key, s));
}

/** Get item at specified position.
    \param index position, [0, getSize()).
    \returns iterator, or end() if index too big. */
StringList::iterator
StringList::getItem(int index) const
{
    iterator i = data.begin();
    while (index > 0 && i != data.end())
        ++i, --index;
    return i;
}

/** Get item by key.
    \returns iterator, end() if not found.
    \post ret != end() <=> ret->first == key */
StringList::iterator
StringList::getItemByKey(int key) const
{
    iterator i = data.begin();
    while (i != data.end() && i->first != key)
        ++i;
    return i;
}

/** Get item by key. Like getItemByKey, but this one returns an index,
    not an iterator.
    \returns index, -1 if not found.
    \post ret != -1 <=> getItem(ret)->first == key */
int
StringList::getItemIndexByKey(int key) const
{
    iterator i = data.begin();
    int n = 0;
    while (i != data.end() && i->first != key)
        ++i, ++n;
    return (i == data.end() ? -1 : n);
}

static bool
sortFunction(const std::pair<int,string_t>& a, const std::pair<int,string_t>& b)
{
    return a.second < b.second;
}

/** Sort contents alphabetically. */
void
StringList::sortAlphabetically()
{
    data.sort(sortFunction);
}

string_t
StringList::get(int index) const
{
    // FIXME: should cache index!
    iterator i = getItem(index);
    if (i != end())
        return i->second;
    else
        return string_t();
}

int
StringList::getLastIndex() const
{
    return data.size();
}
